from collections import OrderedDict
from copy import  deepcopy

#---------------------
#    State space
# --------------------

STATE_DESCRIPTIONS     = OrderedDict( {'v':'vulnerable',
                                       'i':'infected',
                                       's':'symptomatic',
                                       'p':'positive',
                                       'r':'recovered',
                                       'd':'deceased'} )

STATES    = list(STATE_DESCRIPTIONS.keys())
VULNERABLE, INFECTED, SYMPTOMATIC, POSITIVE, RECOVERED, DECEASED = 0, 1, 2, 3, 4, 5     # FIXME adopt from STATE_DESCRIPTIONS
HOME, WORK = 0, 1

#---------------------
#    Parameter space
# --------------------

MOTION_DESCRIPTIONS    = OrderedDict({'t':'Number of time steps in a day',
                                      'k':'Fractional distance moved towards attractor per day (kappa)',
                                      'w':'Standard deviation per day (brownian motion term)'
                                      })

GEOMETRY_DESCRIPTIONS  = OrderedDict({'n':'Population count',
                                      'i':'Number initially infected',
                                      'r':'Radius',
                                      'b':'Bound on size of the world',
                                      'h':'Size of household',
                                      'c':'Commuting fraction',
                                      's':'Spawl',
                                      'e':'Sprawl quadratic term'} )

HEALTH_DESCRIPTIONS    = OrderedDict({ 'vi':'Infection probability if exposed',
                                       'is':'Symptom rate if infected',
                                       'ip':'Randomized testing rate',
                                       'sp':'Symptomatic testing rate',
                                       'ir':'Asymptomatic recovery rate',
                                       'sr':'Symptomatic recovery rate',
                                       'sd':'Symptomatic death rate',
                                       'id':'Asymptomatic death rate',
                                       'pd':'Positive death rate',
                                       'pr':'Positive recovery rate'})  # All rates are per day.

DESCRIPTIONS = OrderedDict({'geometry':GEOMETRY_DESCRIPTIONS,
                             'motion':MOTION_DESCRIPTIONS,
                             'health':HEALTH_DESCRIPTIONS})
CATEGORIES = list(DESCRIPTIONS.keys())
GEOMETRY   = list(GEOMETRY_DESCRIPTIONS.keys())
MOTION     = list(MOTION_DESCRIPTIONS.keys())
HEALTH     = list(HEALTH_DESCRIPTIONS.keys())

NUM_STATES, NUM_GEOMETRY, NUM_MOTION, NUM_HEALTH  = len(STATE_DESCRIPTIONS), len(GEOMETRY_DESCRIPTIONS), len(MOTION), len(HEALTH)




#--------------------------------
#   Map from parameters to R^n
# -------------------------------

# Canonical flattenning of parameters. Not required for simulation but may be useful for estimation, et cetera.

def params_to_vector(d:dict):  # FIXME: Broken ... should adopt ordering from DESCRIPTIONS
    g, m, c  = OrderedDict( zip(GEOMETRY,GEOMETRY) ), OrderedDict( zip(MOTION, MOTION)), OrderedDict( zip(HEALTH, HEALTH) )
    g.update(d['geometry'] )
    m.update(d['motion'])
    c.update(d['health'])
    values = list( g.values() ) + list( m.values() ) + list( c.values() )
    return [ float(v) for v in values ]   # Just checking

def vector_to_params(v):
    return {'geometry':dict(zip(GEOMETRY,v[:NUM_GEOMETRY])),
            'motion':dict(zip(MOTION,v[NUM_GEOMETRY:NUM_GEOMETRY+NUM_MOTION])),
            'health':dict(zip(HEALTH, v[-NUM_STATES * NUM_STATES:]))
           }




from pandemic.example_parameters import LARGE_TOWN
from pandemic.simulation import simulate
import matplotlib.pyplot as plt

# See also /examples

def large_town(with_plot=True):
    simulate(params=LARGE_TOWN, plt=plt if with_plot else None)

def small_town(with_plot=True):
    params = LARGE_TOWN
    params['geometry']['n']= int( params['geometry']['n'] / 2 )
    simulate(params=params, plt=plt if with_plot else None)

def town_that_tests_symptomatic_more(with_plot=True):
    params = LARGE_TOWN
    params['health']['sp']= 3*params['health']['sp']
    simulate(params=params, plt=plt if with_plot else None)

def town_that_tests_randomly(with_plot=True):
    params = LARGE_TOWN
    params['health']['sp']= 3*params['health']['sp']
    params['health']['ip'] =6*params['health']['ip']
    simulate(params=params, plt=plt if with_plot else None)

def town_with_close_neighbours(with_plot=True):
    params = LARGE_TOWN
    params['geometry']['r']= 0.5*params['geometry']['r']   # Sprawl distance
    simulate(params=params, plt=plt if with_plot else None)


if __name__=="__main__":
    large_town()



import pymorton, math

# We use a different z-order embedding for parameters as we need more than two dimensions

def morton_scale(dim):
    return 2 ** 10

def morton_large(dim):
    SCALE = morton_scale(dim=dim)
    return pymorton.interleave(*[SCALE - 1 for _ in range(dim)])

def to_zcurve(prctls):
    """ A mapping from R^n -> R based on the Morton z-curve """
    dim = len(prctls)
    SCALE = morton_scale(dim)
    int_prctls = [int(math.floor(p * SCALE)) for p in prctls]
    m1 = pymorton.interleave(*int_prctls)
    int_prctls_back = pymorton.deinterleave2(m1) if dim == 2 else pymorton.deinterleave3(m1)
    assert all(i1 == i2 for i1, i2 in zip(int_prctls, int_prctls_back))
    m2 = pymorton.interleave(*[SCALE - 1 for _ in range(dim)])
    return m1 / m2

def from_zcurve(zpercentile, dim):
    SCALE = morton_scale(dim)
    zmorton = int(morton_large(dim) * zpercentile + 0.5)
    if dim == 2:
        values = pymorton.deinterleave2(zmorton)
    elif dim == 3:
        values = pymorton.deinterleave3(zmorton)
    prtcls = [v / SCALE for v in values]
    return prtcls



def norminv(self,p):
    f = _norminv_function()
    return f(p)

def to_zscores(prctls):
    norminv = _norminv_function()
    return [ norminv(p) for p in prctls ]

def _norminv_function():
    try:
        from statistics import NormalDist
        return NormalDist(mu=0, sigma=1.0).inv_cdf
    except ImportError:
        from scipy.stats import norm
        return norm.ppf


from pandemic.conventions import STATE_DESCRIPTIONS
from collections import Counter
from pandemic.example_parameters import TOY_TOWN


def plot_callback( positions, status, params, day, day_fraction, home, work, plt, xlabel, step_no, plot_hourly ):
    if plt and (plot_hourly or step_no % 12 == 0):
        plt.clf()
        plot_points(plt=plt, positions=positions, status=status,
                    title="Day " + str(day) + ':' + str(int(100*day_fraction)/100).zfill(4))
        b = params['geometry']['b']
        plt.axis([-b, b, -b, b])
        if xlabel:
            plt.xlabel(xlabel)
        plt.show(block=False)
        plt.pause(0.01)


def plot_points(plt, positions, status, title=None, sizes=None):
    x = [p[0] for p in positions]
    y = [p[1] for p in positions]
    if status is not None:
        if sizes is None:
            if len(positions)>TOY_TOWN['geometry']['n']:
                sizes = [ 1,  16,  9,  9, 4,  16 ]
            else:
                sizes = [ 100, 100, 100, 100, 100, 100]
        scatter = plt.scatter(x=x, y=y, c=status, alpha=0.7, s=[ sizes[s]+4 for s in status ] )
        running = Counter(status)
        descriptions = list(STATE_DESCRIPTIONS.values())
        running_totals  = [ (descriptions[k]+ " ("+str(v)+")",k) for k,v in running.items() ]
        running_totals.sort( key=lambda rt: rt[1] )
        labels = [ desc for desc,_ in running_totals ]
        plt.legend(handles=scatter.legend_elements()[0], labels=labels, loc='upper left')
        if title:
            plt.title(title)
    else:
        scatter = plt.scatter(x=x, y=y)
    return scatter




import random, math
import numpy as np

def sprawl( geometry_params, num ):
    """ Chinese restaurant inspired sprawl model """
    r = geometry_params['r']
    s = geometry_params['s']
    e = geometry_params['e']
    b = geometry_params['b']

    def bound(pos,b):
        return ( (pos[0]+b) % (2*b) - b , (pos[1]+b) % (2*b)  - b )

    def _neighbour(r, s, e, p, b ):
        """ Generate the next location, near to the position p
        :param r:   float          Typical distance to neighbour not including sprawl
        :param s:   float          Sprawl coef
        :param e:   float          Sprawl quadratic term
        :param p:   (float,float)  Neighbour
        :param b:   float          Bound
        :return: (float,float)
        """
        return bound( ( r*np.random.randn()+(1+s)*p[0]+e*p[0]*abs(p[0]) ,
                        r*np.random.randn()+(1+s)*p[1]+e*p[1]*abs(p[1])
               ),b=b)

    points = [ (0,0) ]
    for i in range(num-1):
        p = random.choice(points)
        points.append(_neighbour(r=r,s=s,e=e, p=p, b=b) )
    return points


def home_and_work_locations( geometry_params, num, centers=None ):

    def random_household_size(h):
        return 1 + np.random.binomial(n=6,p=(h-1)/8.0)

    if centers is None:
        b = geometry_params['b']
        centers = [ (b/2*np.random.rand(),b/2*np.random.rand()),(0,-b*np.random.rand())]

    # Sprawl work locations around centers
    work_sprawls = list()
    for center in centers:
        work_sprawls.append( [ (pos[0]+center[0], pos[1]+center[1]) for pos in sprawl( geometry_params=geometry_params, num=num) ] )
    work = [ random.choice(ws) for ws in zip( *work_sprawls ) ]

    # Sprawl homes away from centers also .. though further away
    geometry_params['r'] = 4 * geometry_params['r']
    home_sprawls = list()
    for center in centers:
        home_sprawls.append( [(pos[0] + center[0], pos[1] + center[1]) for pos in sprawl(geometry_params=geometry_params, num=num)])
    home_locations = [random.choice(hs) for hs in zip(*home_sprawls)]
    random.shuffle(home_locations)

    # Squeeze h people in each home.
    h = geometry_params['h']     # Avg number in household
    num_homes = int( math.ceil( num+500 / h ) )
    home = [ hl for hl in home_locations[:num_homes] for _ in range(random_household_size(h)) ][:num]

    # Some stay home
    c = geometry_params['c']
    work = [ w if np.random.rand()<c else h for w,h in zip(work,home) ]
    return home, work

if __name__=="__main__":
    import matplotlib.pyplot as plt
    for _ in range(20):
        plt.close()
        import time
        from pandemic.example_parameters import LARGE_TOWN
        from pandemic.plotting import plot_points
        city = sprawl(geometry_params=LARGE_TOWN['geometry'], num=50000)
        plot_points(plt, city, status=None)
        plt.axis([-20,20,-20,20])
        plt.show()
        plt.pause(0.001)
        time.sleep(1)



from pandemic.movement import nudge, times_of_day
from pandemic.city import home_and_work_locations
from pandemic.conventions import INFECTED, VULNERABLE, POSITIVE, STATE_DESCRIPTIONS, SYMPTOMATIC, DECEASED
from pandemic.compliance import destinations
from pandemic.movement import evolve_positions, newly_exposed
from pandemic.health import contact_progression, individual_progression
from pandemic.plotting import plot_points, plot_callback
from collections import Counter
from pprint import  pprint
import numpy as np



def simulate(params, plt=None, plot_hourly=None, xlabel=None, callback=plot_callback ):
    """ OU Pandemic simulation
    :param params:       dict of dict as per pandemic.conventions
    :param plt:          Handle to matplotlib plot
    :param plot_hourly:  Bool        Set False to speed up, True to see commuting
    :param xlabel:       str         Label for plot
    :param callback:     Any function taking home, work, day, params, positions, status (e.g. for plotting, saving etc)
    :return: None        Use the callback
    """

    if plot_hourly is None:
        plot_hourly = params['geometry']['n']<50000  # Hack, remove

    # Initialize a city's geography and its denizens
    num, num_initially_infected = int(params['geometry']['n']),int(params['geometry']['i'])
    num_times_of_day = int(params['motion']['t'])
    precision  = int(params['geometry']['p'])
    home, work = home_and_work_locations(geometry_params=params['geometry'],num=num)
    positions  = nudge(home,w=0.05*params['motion']['w'])
    status     = np.random.permutation([INFECTED]*num_initially_infected +[VULNERABLE]*(num-num_initially_infected))
    time_step  = 1.0/num_times_of_day

    # Population drifts to work and back, incurring viral load based on proximity to others who are infected
    day = 0
    while any( s in [ INFECTED ] for s in status ):
        day = day+1
        for step_no, day_fraction in enumerate(times_of_day(num_times_of_day)):
            stationary = [ s in [DECEASED, POSITIVE] for s in status ]
            attractors = destinations(status, day_fraction, home, work)
            positions  = evolve_positions(positions=positions, motion_params=params['motion'], attractors=attractors,
                                          time_step=time_step , stationary=stationary )
            exposed = newly_exposed(positions=positions, status=status, precision=precision)
            status = contact_progression(status=status, health_params=params['health'], exposed=exposed)
            status = individual_progression(status, health_params=params['health'], day_fraction=time_step )

            if callback:
                callback(day=day, day_fraction=day_fraction , home=home, work=work, positions=positions, status=status, params=params, step_no=step_no, plot_hourly=plot_hourly, plt=plt)
    pprint(Counter([list(STATE_DESCRIPTIONS.values())[s] for s in status]))




import itertools

# Utils that have nothing to do with the model

def flatten(list_of_list):
    return list(itertools.chain.from_iterable(list_of_list))

def chunks(l, n):
    n = max(1, n)
    return list(l[i:i+n] for i in range(0, len(l), n))


# Entry point for command line. Supply baseline and parameter multipliers in triples
# python3 shell.py large_town geometry n 0.5 health vi 0.5

from pandemic.example_parameters import BASELINES, modifier
from pandemic.conventions import DESCRIPTIONS, CATEGORIES
from pandemic.simulation import simulate
import  matplotlib.pyplot as plt
import sys

def modify_and_run(baseline, triples):
    # python3 shell.py large_town geometry n 20000 health vi 0.5
    params = BASELINES[baseline]
    descriptions = list()
    n = len(triples)
    assert n % 3 ==0, 'Expecting triples of command line parameters '
    num = int(n/3)
    print(num)
    for k in range(num):
        category = triples[3 * k].lower()
        param    = triples[3 * k+1].lower()
        assert category in CATEGORIES
        assert param in list(DESCRIPTIONS[category].keys())
        factor   = float(triples[3*k+2])
        params, desc = modifier( category=category, param=param, factor=factor, baseline=params )
        descriptions.append(desc)
    simulate(params=params, plt=plt, xlabel=','.join(descriptions))

if __name__=="__main__":
    modify_and_run(baseline=sys.argv[1],triples=sys.argv[2:])


import numpy as np
import math
from geohash import encode
from pandemic.conventions import INFECTED, SYMPTOMATIC, POSITIVE, VULNERABLE


def times_of_day(n):
    return [ 1.0*k/n for k in range(n) ]

def nudge(positions,w):
    return [ (pos[0] + w*np.random.randn(),pos[1] + w*np.random.randn()) for pos in positions ]

def random_location(r):
    return (np.random.uniform(-r,r),np.random.uniform(-r,r))

def evolve_positions(positions, motion_params, attractors, time_step, stationary):
    """ OU process time increment """
    W     = motion_params['w'] * math.sqrt(time_step) / math.sqrt(2)
    kappa = motion_params['k'] * time_step
    return[ p if st else (p[0]+W*np.random.randn()+kappa*(a[0]-p[0]),p[1]+W*np.random.randn()+kappa*(a[1]-p[1])) for p,a,st in zip(positions, attractors, stationary) ]

def newly_exposed(positions, status, precision):
    """
         :precision  int      Geohash precision.
         :returns  [ bool ]
    """
    DANGEROUS = set([INFECTED,SYMPTOMATIC])  # Add POSITIVE here if you have less faith in humanity
    infected_locations = set( [ encode(pos[0],pos[1],precision=precision) for pos,s in zip(positions,status) if s in DANGEROUS  ] )
    return [ s==VULNERABLE and encode(pos[0],pos[1],precision=precision) in infected_locations for s,pos in zip(status,positions) ]





# NOT A REAL CONFIG
# Should you want to play around with your own REDIS instance,
# you config might look something like this
REDIS_CONFIG = {"password":"HqtnotarealkeyWkeF",
                "host":"redis-someting-blah.com",
                "port":"35406"}



from pandemic.conventions import HEALTH_DESCRIPTIONS, HEALTH, STATES, VULNERABLE, INFECTED, SYMPTOMATIC, RECOVERED, DECEASED, POSITIVE
from pandemic.util import flatten
import numpy as np


def contact_progression(status, health_params, exposed):
    """ Sometimes exposure leads to infection
    :param exposed:         [ boolean ]
    :return: new status
    """
    return [INFECTED if x and s == VULNERABLE and np.random.randn() < health_params['vi'] else s for s, x in zip(status, exposed)]


def individual_progression(status, health_params, day_fraction):
    """
        :param status         [ int ]     Vector of status
        :param health_params  dict of illness transition parameters
        :param day_fraction   float       Between 0 and 1
    """

    def symptom_emergence(status, p_symptomatic):
        return [ SYMPTOMATIC if s == INFECTED and np.random.rand() < day_fraction*p_symptomatic else s for s in status ]

    def recovery(status, cohort, p_recovery):
        return [RECOVERED if s == cohort and np.random.rand() < day_fraction*p_recovery else s for s in status]

    def death(status, cohort, p_death):
        return [DECEASED if s == cohort and np.random.rand() < day_fraction * p_death else s for s in status]

    def random_testing(status, p_asymptomatic):
        return [POSITIVE if s==INFECTED and np.random.rand() < day_fraction*p_asymptomatic else s for s in status]

    def testing(status, p_symptomatic):
        return [POSITIVE if s==SYMPTOMATIC and np.random.rand() < day_fraction * p_symptomatic else s for s in status]

    status = symptom_emergence(status, p_symptomatic=health_params['is'])
    status = random_testing( status=status, p_asymptomatic=health_params['ip'])
    status = testing(status=status, p_symptomatic=health_params['sp'])
    status = recovery(status, cohort=INFECTED, p_recovery=health_params['ir'])
    status = recovery(status, cohort=POSITIVE, p_recovery=health_params['pr'])
    status = recovery(status, cohort=SYMPTOMATIC, p_recovery=health_params['sr'])
    status = death(status, cohort=INFECTED, p_death=health_params['id'])
    status = death(status, cohort=POSITIVE, p_death=health_params['pd'])
    status = death(status, cohort=SYMPTOMATIC, p_death=health_params['sd'])

    return status


# Aside: useful for tests against matrix status updates ...
CANONICAL_ILLNESS_ORDERING   = flatten([[s1 + s2 for s1 in STATES if s1 + s2 in HEALTH_DESCRIPTIONS.keys()] for s2 in STATES])  # Canonical ordering of non-zero transition parameters


from pandemic.conventions import SYMPTOMATIC, POSITIVE

def destinations(status, day_fraction, home_locations, work_locations, compliant=True):
    """ If sick, stay home! """  # TODO: cohorts for compliance rather than all or nothing.
    return [h if day_fraction >= 0.5 or (compliant and s in [SYMPTOMATIC, POSITIVE]) else w for s, h, w in zip(status, home_locations, work_locations)]



# A community experiment to build a surrogate model

import requests, math, copy, json, random
from collections import Counter
from pandemic.conventions import STATES
from pandemic.zcurves import to_zcurve, from_zcurve
from pandemic.example_parameters import BASELINES
from pandemic.conventions import params_to_vector, vector_to_params, CATEGORIES, STATE_DESCRIPTIONS
from pandemic.zcurves import to_zcurve
from pandemic.simulation import simulate
from pandemic.plotting import plot_points
import numpy as np
from pprint import pprint
import matplotlib.pyplot as plt

#-------------------------
#   Free parameters
#-------------------------

FREE_PARAMS = {'motion':{'w'},
               'geometry':{'c'},
               'health':{'sp'}}
FREE_DIM = len(FREE_PARAMS)

def free_param_percentiles(params,baseline='large_town'):
    """ Takes parameter set and returns a vector of numbers between 0 and 1 """
    free = list()
    for c in CATEGORIES:
        for ky in FREE_PARAMS[c]:
            raw = params[c][ky]
            default = BASELINES[baseline][c][ky]
            ratio   = raw / default
            prctl = 1. / (1. + math.exp(-ratio)/math.exp(-1))
            free.append(prctl)
    return free

PARAMS_SCALE = 1000 * 1000 * 1000 * 1000 * 1000
def cube_to_int(v):
    """ z=curve """
    assert 2<=len(v)<=3
    assert [ abs(v_)<=1.0 for v_ in v]
    z = to_zcurve(v)
    return int(PARAMS_SCALE * z)

def int_to_cube(h, dim):
    """ Map integer back to the cube"""
    z = h / PARAMS_SCALE
    v = from_zcurve(z,dim)
    return v

DAY_SCALE = 10000
def days_to_int(day, day_fraction):
    return int(DAY_SCALE*(day+day_fraction))

def int_to_days(key):
    return int(key) / DAY_SCALE


#----------------------------------------
#   Examples of reporting callbacks
#----------------------------------------

def health_metric(status, **ignore_other_kwargs):
    counts = Counter(status)
    return [ counts[k] for k,_ in enumerate(STATES) ]

def random_modification_ratio():
    return math.exp(np.random.randn())

def random_modification(params):
    for c, prms in FREE_PARAMS.items():
        for prm in list(prms):
            params[c][prm] = random_modification_ratio()*params[c][prm]
    return params

#----------------------------------------
#   Client for www.swarmprediction.com
#----------------------------------------

class Surrogate():

    def __init__(self, baseline, url='https://www.swarmprediction.com/metrics', callback_metric=health_metric, plt=None, params=None, quietude=24):
        self.baseurl  = url
        self.baseline = baseline
        self.callback_metric = callback_metric
        self.params   = params or random_modification( copy.deepcopy( BASELINES[baseline] ) )
        self.WIDTH    = 2+len(str(DAY_SCALE))+len(str(PARAMS_SCALE))
        self.time_history = list()
        self.key_history  = list()
        self.metric_history = list()
        self.quietude = quietude
        self.plt = plt
        if plt is not None:
            self.fig, self.axs = plt.subplots(nrows=2,ncols=2)

    def to_key(self, day, day_fraction ):
        return str(self.to_int(day=day, day_fraction=day_fraction, params=self.params)).zfill(self.WIDTH)

    def to_int(self, day, day_fraction, params ):
        """ Map (time,params) to integer """
        v = free_param_percentiles(params)
        params_as_int = cube_to_int(v)
        time_as_int = days_to_int(day=day, day_fraction=day_fraction)
        return time_as_int*PARAMS_SCALE + params_as_int

    def from_int(self,time_and_params):
        """ Map integer back to (time,params) """
        elapsed      = int( math.floor( time_and_params / PARAMS_SCALE ) )
        params_as_int = time_and_params % PARAMS_SCALE
        params = int_to_cube(params_as_int,dim=FREE_DIM)
        return elapsed, params

    def run(self):
        simulate(params=self.params,callback=self.callback,plot_hourly=False,plt=self.plt,xlabel="Sending results to www.swarmprediction.com. Thanks!")

    def callback(self, day, day_fraction, status, positions, home, work, plt, **ignore_other_kwargs):
        """ This gets called after each computation cycle (see pandemic/simulation.py) """
        if day_fraction==0:
            metrics = self.callback_metric(status=status)
            ky      = self.to_key(day=day, day_fraction=day_fraction)
            res     = self.post(key=ky, metrics=metrics)
            self.key_history.append(ky)
            self.metric_history.append(metrics)
            self.time_history.append(day+day_fraction)
            if random.choice(range(self.quietude)) == 0:
                if self.plt is not None:
                    self.plot(plt=plt,positions=positions,status=status)
                pprint({"key":ky,"day":str(day+day_fraction),"result":res,"retrieve":self.baseurl.replace('metrics',self.baseline)+'/'+ky})
                print(' ',flush=True)

    def post(self, key, metrics):
        """ The server stores results in a REDIS sorted set, where the score is an embedding of the parameters
            from R^3 into R via a Morton curve. Similar simulations with nearby parameters can be retrieved if
            they are close on the space filling curve ... which is often though not always the case. We *may* introduce
            redundancy into the embeddings to improve this somewhat.
        """
        res = requests.post(self.baseurl +'/' + key, data={'baseline':self.baseline,'metrics':json.dumps(metrics)})
        return res.status_code if res.status_code not in [200,201] else res.json()

    def get(self,key):
        res = requests.get(url=self.baseurl + '/' + str(key), params={'baseline':self.baseline} )
        return res.status_code if res.status_code not in [200,201] else res.json()

    def get_nearby(self, day, day_fraction, params, precision=6):
        """ Retrieves scenarios which used similar parameters and the same exact time step """
        key  = self.to_key(day=day, day_fraction=day_fraction, params=params)
        res  = requests.patch(url=self.baseurl+'/' + str(key), data={'precision':precision,'baseline':self.baseline})
        return res.json() if res.status_code==200 else res.status_code

    def plot(self,plt,positions,status):

        # Population plot
        self.axs[0][0].clear()
        plot_points( plt=self.axs[0][0], positions=positions, status=status )
        self.axs[0][0].figure

        # Metrics plots, regular and logarithmic
        for k in range(2):
            self.axs[1][k].clear()
            self.plot_metrics(plt=self.axs[1][k], logarithmic=k)
            self.axs[1][k].figure

        # Rates of change
        self.axs[0][1].clear()
        self.plot_metrics(plt=self.axs[0][1], logarithmic=False, differences=True)
        self.axs[0][1].figure


        plt.show(block=False)
        plt.pause(0.01)

    def plot_metrics(self, plt, logarithmic, differences=False):
        metrics = list(zip(*self.metric_history))[1:]
        if len(metrics[0])>3:
            for m in metrics:
                if differences:
                    plt.plot(self.time_history[1:], list(np.diff(m)))
                    plt.set_ylabel('Daily change')
                else:
                    plt.plot(self.time_history,m)
                    plt.set_ylabel('Cumulative')
            if logarithmic:
                plt.set_yscale('log')
            labels = list(STATE_DESCRIPTIONS.values())[1:]
            if differences:
                labels = [ 'net newly '+lb for lb in labels ]
            plt.legend(labels)
            plt.set_xlabel('Days since first '+str(self.params['geometry']['i'])+' infections.')

def surrogate(baseline='city',plot=True,quietude=5):
    print('Starting pandemic simulation',flush=True)
    if plot:
        try:
            import matplotlib.pyplot as plt
        except:
            plt = None
            print("Plotting does not seem to be working, but calculations will proceed")
    else:
        plt=None

    s = Surrogate(plt=plt,baseline=baseline,quietude=quietude)
    pprint(s.params)
    s.run()
    if s.plt is not None:
        s.plt.close()
    while True:
        s = Surrogate(plt=None,baseline=baseline,quietude=quietude)
        pprint(s.params)
        s.run()

if __name__=="__main__":
    surrogate(baseline='town',quietude=1)
